#!/usr/bin/env python3

import re

# Plural and singular units
UNITS = {
    "bags": "bag",
    "bars": "bar",
    "baskets": "basket",
    "batches": "batch",
    "blocks": "block",
    "bottles": "bottle",
    "boxes": "box",
    "branches": "branch",
    "bulbs": "bulb",
    "bunches": "bunch",
    "bundles": "bundle",
    "cans": "can",
    "chops": "chop",
    "chunks": "chunk",
    "cloves": "clove",
    "clusters": "cluster",
    "cm": "cm",
    "cubes": "cube",
    "cups": "cup",
    "cutlets": "cutlet",
    "dashes": "dash",
    "dollops": "dollop",
    "drops": "drop",
    "ears": "ear",
    "envelopes": "envelope",
    "feet": "foot",
    "fillets": "fillet",
    "g": "g",
    "gallons": "gallon",
    "glasses": "glass",
    "grams": "gram",
    "grinds": "grind",
    "handfuls": "handful",
    "heads": "head",
    "inches": "inch",
    "jars": "jar",
    "kg": "kg",
    "kilograms": "kilogram",
    "knobs": "knob",
    "lbs": "lb",
    "leaves": "leaf",
    "lengths": "length",
    "links": "link",
    "l": "l",
    "liters": "liter",
    "litres": "litre",
    "loaves": "loaf",
    "milliliters": "milliliter",
    "ml": "ml",
    "mugs": "mug",
    "ounces": "ounce",
    "oz": "oz",
    "packs": "pack",
    "packages": "package",
    "packets": "packet",
    "pairs": "pair",
    "pieces": "piece",
    "pinches": "pinch",
    "pints": "pint",
    "pods": "pod",
    "pounds": "pound",
    "racks": "rack",
    "rashers": "rasher",
    "recipes": "recipe",
    "rectangles": "rectangle",
    "ribs": "rib",
    "quarts": "quart",
    "scoops": "scoop",
    "segments": "segment",
    "shakes": "shake",
    "sheets": "sheet",
    "shots": "shot",
    "shoots": "shoot",
    "slabs": "slab",
    "slices": "slice",
    "sprigs": "sprig",
    "squares": "square",
    "stalks": "stalk",
    "steaks": "steak",
    "stems": "stem",
    "sticks": "stick",
    "strips": "strip",
    "tablespoons": "tablespoon",
    "tbsps": "tbsp",
    "tbs": "tb",
    "teaspoons": "teaspoon",
    "tsps": "tsp",
    "twists": "twist",
    "wedges": "wedge",
    "wheels": "wheel",
}
# Generate capitalized version of each entry in the UNITS dictionary
_capitalized_units = {}
for plural, singular in UNITS.items():
    _capitalized_units[plural.capitalize()] = singular.capitalize()
UNITS = UNITS | _capitalized_units

# Words that can modify a unit
UNIT_MODIFIERS = [
    "big",
    "fat",
    "generous",
    "healthy",
    "heaped",
    "heaping",
    "large",
    "medium",
    "medium-size",
    "medium-sized",
    "scant",
    "small",
    "thick",
    "thin",
]

# Units that can be part of the name
# e.g. 1 teaspoon ground cloves, or 5 bay leaves
AMBIGUOUS_UNITS = [
    "cloves",
    "leaves",
    "slabs",
    "wedges",
]
# Extend list automatically to include singular and capitalized forms
_ambiguous_units_alt_forms = []
for amb_unit in AMBIGUOUS_UNITS:
    _ambiguous_units_alt_forms.append(amb_unit.capitalize())
    _ambiguous_units_alt_forms.append(UNITS[amb_unit])
    _ambiguous_units_alt_forms.append(UNITS[amb_unit.capitalize()])

AMBIGUOUS_UNITS.extend(_ambiguous_units_alt_forms)


# Strings and their numeric representation
STRING_NUMBERS = {
    "one": "1",
    "two": "2",
    "three": "3",
    "four": "4",
    "five": "5",
    "six": "6",
    "seven": "7",
    "eight": "8",
    "nine": "9",
    "ten": "10",
    "eleven": "11",
    "twelve": "12",
    "thirteen": "13",
    "fourteen": "14",
    "fifteen": "15",
    "sixteen": "16",
    "seventeen": "17",
    "eighteen": "18",
    "nineteen": "19",
}
# Precompile the regular expressions for matching the string numbers
STRING_NUMBERS_REGEXES = {}
for s, n in STRING_NUMBERS.items():
    # This is case insensitive so it replace e.g. "one" and "One"
    # Only match if the string is preceded by a non-word character or is at
    # the start of the sentence
    STRING_NUMBERS_REGEXES[s] = (re.compile(rf"\b({s})\b", flags=re.IGNORECASE), n)



#!/usr/bin/env python3

import re
from fractions import Fraction
from html import unescape
from typing import Any

from nltk.stem.porter import PorterStemmer
from nltk.tag import pos_tag
from nltk.tokenize import RegexpTokenizer

from ._constants import AMBIGUOUS_UNITS, STRING_NUMBERS_REGEXES, UNITS

# Regex pattern for fraction parts.
# Matches 0+ numbers followed by 0+ white space characters followed by a number then
# a forward slash then another number.
FRACTION_PARTS_PATTERN = re.compile(r"(\d*\s*\d/\d+)")

# Regex pattern for checking if token starts with a capital letter.
CAPITALISED_PATTERN = re.compile(r"^[A-Z]")

# Regex pattern for finding quantity and units without space between them.
# Assumes the quantity is always a number and the units always a letter.
QUANTITY_UNITS_PATTERN = re.compile(r"(\d)\-?([a-zA-Z])")

# Regex pattern for matching a numeric range e.g. 1-2, 2-3.
RANGE_PATTERN = re.compile(r"\d+\s*[\-]\d+")

# Regex pattern for matching a range in string format e.g. 1 to 2, 8.5 to 12, 4 or 5.
# Assumes fake fractions and unicode fraction have already been replaced.
# Allows the range to include a hyphen, which are captured in separate groups.
# Captures the two number in the range in separate capture groups.
STRING_RANGE_PATTERN = re.compile(r"([\d\.]+)(\-)?\s*(to|or)\s*(\-)*\s*([\d\.]+(\-)?)")

# Regex pattern to match hyphen followed by a number. This is used to fix ranges
# following the replacement of unicode fractions with fake fractions.
# The pattern any characters up to a hyphen followed by a space followed by a number.
# The parts before and after that space are cpatured in groups so we can reconstitute
# the sentence without that middle space.
BROKEN_RANGE_PATTERN = re.compile(r"(.*\-)\s(\d.*)")

# Define tokenizer.
# We are going to split an sentence between substrings that match the following groups
# a) letters and any punctuation, except
# b) open and close parentheses, open and close brackets, open and close braces,
# quote, comma.
# The following punctuation is deliberately left out of the these groups so that
# they are removed: backslash.
group_a = r"[\w!\#\$\£\€%\&'\*\+\-\.>=<\?@\^_`\\\|\~’]+"
group_b = r"[\(\)\[\]\{\}\,\"/:;]"
REGEXP_TOKENIZER = RegexpTokenizer(rf"{group_a}|{group_b}", gaps=False)

STEMMER = PorterStemmer()


class PreProcessor:

    """Recipe ingredient sentence PreProcessor class.

    Performs the necessary preprocessing on a sentence to generate the features
    required for the ingredient parser model.

    Each input sentence goes through a cleaning process to tidy up the input into a
    standardised form.

    Notes
    -----

    The cleaning steps are as follows

    1. | Replace all en-dashes and em-dashes with hyphens.
    2. | Replace numbers given as words with the numeric equivalent.
       | e.g. one >> 1
    3. | Replace fractions given in html markup with the unicode representation.
       | e.g. &frac12; >> ½
    4. | Replace unicode fractions with the equivalent decimal form. Decimals are
       | rounded to 3 a maximum of decimal places.
       | e.g. ½ >> 0.5
    5. | Replace "fake" fractions represented by 1/2, 2/3 etc. with the equivalent
       | decimal form
       | e.g. 1/2 >> 0.5
    6. | A space is enforced between quantities and units
    7. | Remove trailing periods from units
       | e.g. tsp. >> tsp
    8. | Numeric ranges indicated in words using "to" or "or" are replaced with a
       | standard numeric form
       | e.g. 1 or 2 >> 1-2; 10 to 12 >> 10-12
    9. | Units are made singular. This step uses a predefined list of plural units and
       | their singular form.

    Following the cleaning of the input sentence, it is tokenized into a list of tokens.

    Each token is one of the following

    * A word, including most punctuation marks
    * Opening or closing parentheses, braces, brackets; comma; speech marks

    The features for each token are computed on demand using the ``sentence_features``
    method, which returns a list of dictionaries.
    Each dictionary is the feature set for each token.

    The sentence features can then be passed to the CRF model which will generate the
    parsed output.

    Parameters
    ----------
    input_sentence : str
        Input ingredient sentence.
    defer_pos_tagging : bool
        Defer part of speech tagging until feature generation.
        Part of speech tagging is an expensive operation and it's not always needed when
        using this class.
    show_debug_output : bool, optional
        If True, print out each stage of the sentence normalisation

    Attributes
    ----------
    defer_pos_tagging : bool
        Defer part of speech tagging until feature generation
    show_debug_output : bool
        If True, print out each stage of the sentence normalisation
    input : str
        Input ingredient sentence.
    pos_tags : list[str]
        Part of speech tag for each token in the tokenized sentence.
    sentence : str
        Input ingredient sentence, cleaned to standardised form.
    singularised_indices : list[int]
        Indices of tokens in tokenised sentence that have been converted from plural
        to singular
    tokenized_sentence : list[str]
        Tokenised ingredient sentence.
    """

    def __init__(
        self,
        input_sentence: str,
        defer_pos_tagging: bool = False,
        show_debug_output: bool = False,
    ):
        """Initialisation

        Parameters
        ----------
        input_sentence : str
            Input ingredient sentence
        defer_pos_tagging : bool, optional
            Defer part of speech tagging until feature generation
        show_debug_output : bool, optional
            If True, print out each stage of the sentence normalisation

        """
        self.show_debug_output = show_debug_output
        self.input: str = input_sentence
        self.sentence: str = self._normalise(input_sentence)

        _tokenised_sentence = REGEXP_TOKENIZER.tokenize(self.sentence)
        (
            self.tokenized_sentence,
            self.singularised_indices,
        ) = self._singlarise_units(_tokenised_sentence)

        self.defer_pos_tagging: bool = defer_pos_tagging
        if not defer_pos_tagging:
            self.pos_tags: list[str] = self._tag_partofspeech(self.tokenized_sentence)
        else:
            self.pos_tags = []

    def __repr__(self) -> str:
        """__repr__ method

        Returns
        -------
        str
            String representation of initialised object
        """
        return f'PreProcessor("{self.input}")'

    def __str__(self) -> str:
        """__str__ method

        Returns
        -------
        str
            Human readable string representation of object
        """
        _str = [
            "Pre-processed recipe ingredient sentence",
            f"\t    Input: {self.input}",
            f"\t  Cleaned: {self.sentence}",
            f"\tTokenized: {self.tokenized_sentence}",
        ]
        return "\n".join(_str)

    def _normalise(self, sentence: str) -> str:
        """Normalise sentence prior to feature extraction

        Parameters
        ----------
        sentence : str
            Ingredient sentence

        Returns
        -------
        str
            Normalised ingredient sentence
        """
        # List of functions to apply to sentence
        # Note that the order matters
        funcs = [
            self._replace_en_em_dash,
            self._replace_string_numbers,
            self._replace_html_fractions,
            self._replace_unicode_fractions,
            self._replace_fake_fractions,
            self._split_quantity_and_units,
            self._remove_unit_trailing_period,
            self._replace_string_range,
        ]

        for func in funcs:
            sentence = func(sentence)

            if self.show_debug_output:
                print(f"{func.__name__}: {sentence}")

        return sentence.strip()

    def _replace_en_em_dash(self, sentence: str) -> str:
        """Replace en-dashes and em-dashes with hyphens.

        Parameters
        ----------
        sentence : str
            Ingredient sentence

        Returns
        -------
        str
            Ingredient sentence with en and em dashes replaced with hyphens

        Examples
        --------
        >>> p = PreProcessor("")
        >>> p._replace_en_em_dash("2 cups flour – white or self-raising")
        "2 cups flour - white or self-raising"

        >>> p = PreProcessor("")
        >>> p._replace_en_em_dash("3–4 sirloin steaks")
        "3-4 sirloin steaks"
        """
        return sentence.replace("–", "-").replace("—", "-")

    def _replace_string_numbers(self, sentence: str) -> str:
        """Replace string numbers (e.g. one, two) with numeric values (e.g. 1, 2)

        Parameters
        ----------
        sentence : str
            Ingredient sentence

        Returns
        -------
        str
            Ingredient sentence with string numbers replace with numeric values

        Examples
        --------
        >>> p = PreProcessor("")
        >>> p._replace_string_numbers("three large onions")
        "3 large onions"

        >>> p = PreProcessor("")
        >>> p._replace_string_numbers("twelve bonbons")
        "12 bonbons"
        """

        # STRING_NUMBER_REGEXES is a dict where the values are a tuple of the compiled
        # regular expression for matching a string number e.g. 'one', 'two' and the
        # substitution numerical value for that string number.
        for regex, substitution in STRING_NUMBERS_REGEXES.values():
            sentence = regex.sub(rf"{substitution}", sentence)

        return sentence

    def _replace_html_fractions(self, sentence: str) -> str:
        """Replace html fractions e.g. &frac12; with unicode equivalents

        Parameters
        ----------
        sentence : str
            Ingredient sentence

        Returns
        -------
        str
            Ingredient sentence with html fractions replaced

        Examples
        --------
        >>> p = PreProcessor("")
        >>> p._replace_html_fractions("1&frac34; cups tomato ketchup")
        "1¾ cups tomato ketchup"
        """
        return unescape(sentence)

    def _replace_fake_fractions(self, sentence: str) -> str:
        """Attempt to parse fractions from sentence and convert to decimal
        This looks for fractions with the format of 1/2, 1/4, 1 1/2 etc.

        Parameters
        ----------
        sentence : str
            Ingredient sentence

        Returns
        -------
        str
            Ingredient sentence with fractions replaced with decimals

        Examples
        --------
        >>> p = PreProcessor("")
        >>> p._replace_fake_fractions("1/2 cup icing sugar")
        "0.5 cup icing sugar"

        >>> p = PreProcessor("")
        >>> p._replace_fake_fractions("2 3/4 pound chickpeas")
        "2.75 pound chickpeas"
        """
        matches = FRACTION_PARTS_PATTERN.findall(sentence)

        if not matches:
            return sentence

        # This is a bit of a hack.
        # If a fraction appears multiple times but in different forms e.g. 1/2 and
        # 1 1/2, then
        # we need to replace the longest one first, otherwise both instance of 1/2
        # would be replaced at the same time which would mean that the instance of
        # 1 1/2 would end up as 1 0.5 instead of 1.5
        # Before we sort, we need to strip any space from the start and end.
        matches = [match.strip() for match in matches]
        matches.sort(key=len, reverse=True)

        for match in matches:
            split = match.split()
            summed = float(sum(Fraction(s) for s in split))
            rounded = round(summed, 3)
            sentence = sentence.replace(match, f"{rounded:g}")

        return sentence

    def _replace_unicode_fractions(self, sentence: str) -> str:
        """Replace unicode fractions with a 'fake' ascii equivalent.
        The ascii equivalent is used because the replace_fake_fractions function can
        deal with spaces between an integer and the fraction.

        Parameters
        ----------
        sentence : str
            Ingredient sentence

        Returns
        -------
        str
            Ingredient sentence with unicode fractions replaced

        Examples
        --------
        >>> p = PreProcessor("")
        >>> p._replace_unicode_fractions("½ cup icing sugar")
        "1/2 cup icing sugar"

        >>> p = PreProcessor("")
        >>> p._replace_unicode_fractions("3⅓ cups warm water")
        "3 1/3 cups warm water"
        """
        fractions = {
            "\u215b": "1/8",
            "\u215c": "3/8",
            "\u215d": "5/8",
            "\u215e": "7/8",
            "\u2159": "1/6",
            "\u215a": "5/6",
            "\u2155": "1/5",
            "\u2156": "2/5",
            "\u2157": "3/5",
            "\u2158": "4/5",
            "\xbc": "1/4",
            "\xbe": "3/4",
            "\u2153": "1/3",
            "\u2154": "2/3",
            "\xbd": "1/2",
        }
        for f_unicode, f_ascii in fractions.items():
            # Insert space before ascii fraction to avoid merging into the
            # previous token.
            sentence = sentence.replace(f_unicode, f" {f_ascii}")

        # Because we inserted the space before the fake fraction, we might have broken
        # some ranges by adding a space after the hyphen, so let's fix that.
        sentence = BROKEN_RANGE_PATTERN.sub(r"\1\2", sentence)

        return sentence

    def _split_quantity_and_units(self, sentence: str) -> str:
        """Insert space between quantity and unit
        This currently finds any instances of a number followed directly by a letter
        with no space in between.

        Parameters
        ----------
        sentence : str
            Ingredient sentence

        Returns
        -------
        str
            Ingredient sentence with spaces inserted between quantity and units

        Examples
        --------
        >>> p = PreProcessor("")
        >>> p._split_quantity_and_units("100g green beans")
        "100 g green beans"

        >>> p = PreProcessor("")
        >>> p._split_quantity_and_units("2-pound red peppers, sliced")
        "2 pound red peppers, sliced"
        """
        return QUANTITY_UNITS_PATTERN.sub(r"\1 \2", sentence)

    def _remove_unit_trailing_period(self, sentence: str) -> str:
        """Remove trailing periods from units e.g. tsp. -> tsp

        Parameters
        ----------
        sentence : str
            Ingredient sentence

        Returns
        -------
        str
            Ingredient sentence with trailing periods from units removed

        Examples
        --------
        >>> p = PreProcessor("")
        >>> p._remove_unit_trailing_period("1 tsp. garlic powder")
        "1 tsp garlic powder"

        >>> p = PreProcessor("")
        >>> p._remove_unit_trailing_period("5 oz. chopped tomatoes")
        "5 oz chopped tomatoes"
        """
        units = ["tsp.", "tsps.", "tbsp.", "tbsps.", "lb.", "lbs.", "oz."]
        units.extend([u.capitalize() for u in units])
        for unit in units:
            unit_no_period = unit.replace(".", "")
            sentence = sentence.replace(unit, unit_no_period)

        return sentence

    def _replace_string_range(self, sentence: str) -> str:
        """Replace range in the form "<num> to <num" with
        standardised range "<num>-<num>".

        For example
        -----------
        1 to 2 -> 1-2
        8.5 to 12.5 -> 8.5-12.5
        16- to 9-

        Parameters
        ----------
        sentence : str
            Ingredient sentence

        Returns
        -------
        str
            Ingredient sentence with string ranges replaced with standardised range

        Examples
        --------
        >>> p = PreProcessor("")
        >>> p._replace_string_range("1 to 2 mashed bananas")
        "1-2 mashed bananas"

        >>> p = PreProcessor("")
        >>> p._replace_string_range("4- or 6- large apples")
        "5-6 large apples"
        """
        return STRING_RANGE_PATTERN.sub(r"\1-\5", sentence)

    def _singlarise_units(
        self, tokenised_sentence: list[str]
    ) -> tuple[list[str], list[int]]:
        """Singularise units in tokenised sentence and return list of singularised
        indices e.g. cups -> cup, tablespoons -> tablespoon

        Parameters
        ----------
        tokenised_sentence : list[str]
            Tokenised sentence

        Returns
        -------
        list[str]
            Tokenised sentence with units singularised
        list[int]
            List of indices of tokenised sentence that have been singularised
        """
        singularised_indices = []
        for idx, token in enumerate(tokenised_sentence):
            singular = UNITS.get(token, None)
            if singular is not None:
                tokenised_sentence[idx] = singular
                singularised_indices.append(idx)

        return (tokenised_sentence, singularised_indices)

    def _tag_partofspeech(self, tokens: list[str]) -> list[str]:
        """Tag tokens with part of speech using universal tagset

        This function manually fixes tags that are incorrect in the context of
        ----------------------------------------------------------------------
        1. Change tags of numeric ranges to CD
        2. Change tag of "ground" from NN to VBD e.g. ground almonds


        Parameters
        ----------
        tokens : list[str]
            Tokenized ingredient sentence

        Returns
        -------
        list[str]
            List of part of speech tags
        """
        tags = []
        # If we don't make each token lower case, that POS tag maybe different in
        # ways that are unhelpful. For example, if a sentence starts with a unit.
        for token, tag in pos_tag([t.lower() for t in tokens]):
            if RANGE_PATTERN.match(token):
                tag = "CD"
            if token in ["ground"]:
                tag = "VBD"
            tags.append(tag)
        return tags

    def _is_unit(self, token: str) -> bool:
        """Return True if token is a unit

        Parameters
        ----------
        token : str
            Token to check

        Returns
        -------
        bool
            True if token is a unit, else False

        Examples
        --------
        >>> p = PreProcessor("")
        >>> p._is_unit("cup")
        True

        >>> p = PreProcessor("")
        >>> p._is_unit("beef")
        False
        """
        return token.lower() in UNITS.values()

    def _is_numeric(self, token: str) -> bool:
        """Return True if token is numeric

        Parameters
        ----------
        token : str
            Token to check

        Returns
        -------
        bool
            True if token is numeric, else False

        Examples
        --------
        >>> p = PreProcessor("")
        >>> p._is_numeric("1")
        True

        >>> p = PreProcessor("")
        >>> p._is_numeric("1-2")
        True

        >>> p = PreProcessor("")
        >>> p._is_numeric("dozen")
        True

        >>> p = PreProcessor("")
        >>> p._is_numeric("beef")
        False
        """
        if "-" in token:
            parts = token.split("-")
            return all([self._is_numeric(part) for part in parts])

        if token == "dozen":
            return True

        try:
            float(token)
            return True
        except ValueError:
            return False

    def _follows_comma(self, index: int) -> bool:
        """Return True if token at index follows a comma (by any amount) in sentence
        If the token at index is a comma, treat it the same as any other token

        Parameters
        ----------
        index : int
            Index of token to check

        Returns
        -------
        bool
            True if token follows comma, else False
        """
        return "," in self.tokenized_sentence[:index]

    def _follows_plus(self, index: int) -> bool:
        """Return True if token at index follow "plus" by any amount in sentence.
        If the token at the index is "plus", it doesn't count as following.

        Parameters
        ----------
        index : int
            Index of token to check

        Returns
        -------
        bool
            True if token follows "plus", else False
        """
        return "plus" in self.tokenized_sentence[:index]

    def _is_capitalised(self, token: str) -> bool:
        """Return True if token starts with a capital letter

        Parameters
        ----------
        token : str
            Token to check

        Returns
        -------
        bool
            True if token starts with a capital letter, else False

        Examples
        --------
        >>> p = PreProcessor("")
        >>> p._is_capitalised("Chicken")
        True

        >>> p = PreProcessor("")
        >>> p._is_capitalised("chicken")
        False
        """
        return CAPITALISED_PATTERN.match(token) is not None

    def _is_inside_parentheses(self, index: int) -> bool:
        """Return True if token is inside parentheses within the sentence or is a
        parenthesis.

        Parameters
        ----------
        index : int
            Index of token to check

        Returns
        -------
        bool
            True if index is inside parentheses or is parenthesis, else False
        """
        # If it's "(" or ")", return True
        if self.tokenized_sentence[index] in ["(", ")", "[", "]"]:
            return True

        open_parens, closed_parens = [], []
        for i, token in enumerate((self.tokenized_sentence)):
            if token == "(" or token == "[":
                open_parens.append(i)
            elif token == ")" or token == "]":
                closed_parens.append(i)

        for start, end in zip(open_parens, closed_parens):
            if start < index < end:
                return True

        return False

    def _is_ambiguous_unit(self, token: str) -> bool:
        """Return True if token is in AMBIGUOUS_UNITS list

        Parameters
        ----------
        token : str
            Token to check

        Returns
        -------
        bool
            True if token is in AMBIGUOUS_UNITS, else False

        Examples
        --------
        >>> p = PreProcessor("")
        >>> p._is_ambiguous_unit("cloves")
        True

        >>> p = PreProcessor("")
        >>> p._is_ambiguous_unit("wedge")
        True

        >>> p = PreProcessor("")
        >>> p._is_ambiguous_unit("leaf")
        True
        """
        return token in AMBIGUOUS_UNITS

    def _token_features(self, index: int) -> dict[str, str | bool]:
        """Return the features for each token in the sentence

        Parameters
        ----------
        index : int
            Index of token to get features for.

        Returns
        -------
        dict[str, Any]
            Dictionary of features for token at index
        """
        token = self.tokenized_sentence[index]
        features = {
            "stem": STEMMER.stem(token),
            "pos": self.pos_tags[index],
            "is_capitalised": self._is_capitalised(token),
            "is_numeric": self._is_numeric(token),
            "is_unit": self._is_unit(token),
            "is_ambiguous": self._is_ambiguous_unit(token),
            "is_in_parens": self._is_inside_parentheses(index),
            "is_after_comma": self._follows_comma(index),
            "is_after_plus": self._follows_plus(index),
        }

        if index > 0:
            features["prev_pos"] = self.pos_tags[index - 1]
            features["prev_word"] = STEMMER.stem(self.tokenized_sentence[index - 1])

        if index > 1:
            features["prev_pos2"] = self.pos_tags[index - 2]
            features["prev_word2"] = STEMMER.stem(self.tokenized_sentence[index - 2])

        if index < len(self.tokenized_sentence) - 1:
            features["next_pos"] = self.pos_tags[index + 1]
            features["next_word"] = STEMMER.stem(self.tokenized_sentence[index + 1])

        if index < len(self.tokenized_sentence) - 2:
            features["next_pos2"] = self.pos_tags[index + 2]
            features["next_word2"] = STEMMER.stem(self.tokenized_sentence[index + 2])

        return features

    def sentence_features(self) -> list[dict[str, Any]]:
        """Return features for all tokens in sentence

        Returns
        -------
        list[dict[str, Any]]
            List of features for each token in sentence
        """
        if self.defer_pos_tagging:
            # If part of speech tagging was deferred, do it now
            self.pos_tags = self._tag_partofspeech(self.tokenized_sentence)

        features = []
        for idx, _ in enumerate(self.tokenized_sentence):
            features.append(self._token_features(idx))

        return features



#!/usr/bin/env python3

from dataclasses import dataclass
from importlib.resources import as_file, files

import pycrfsuite

from ._utils import pluralise_units
from .postprocess import ParsedIngredient, PostProcessor
from .preprocess import PreProcessor

# Create TAGGER object
TAGGER = pycrfsuite.Tagger()
with as_file(files(__package__) / "model.crfsuite") as p:
    TAGGER.open(str(p))


def parse_ingredient(sentence: str) -> ParsedIngredient:
    """Parse an ingredient sentence using CRF model to return structured data

    Parameters
    ----------
    sentence : str
        Ingredient sentence to parse

    Returns
    -------
    ParsedIngredient
        ParsedIngredient object of structured data parsed from input string
    """

    processed_sentence = PreProcessor(sentence)
    tokens = processed_sentence.tokenized_sentence
    labels = TAGGER.tag(processed_sentence.sentence_features())
    scores = [TAGGER.marginal(label, i) for i, label in enumerate(labels)]

    # Re-pluralise tokens that were singularised if the label isn't UNIT
    # For tokens with UNIT label, we'll deal with them below
    for idx in processed_sentence.singularised_indices:
        token = tokens[idx]
        label = labels[idx]
        if label != "UNIT":
            tokens[idx] = pluralise_units(token)

    postprocessed_sentence = PostProcessor(sentence, tokens, labels, scores)
    return postprocessed_sentence.parsed()


def parse_multiple_ingredients(sentences: list[str]) -> list[ParsedIngredient]:
    """Parse multiple ingredient sentences in one go.

    This function accepts a list of sentences, with element of the list representing
    one ingredient sentence.
    A list of dictionaries is returned, with optional confidence values.
    This function is a simple for-loop that iterates through each element of the
    input list.

    Parameters
    ----------
    sentences : list[str]
        List of sentences to parse

    Returns
    -------
    list[ParsedIngredient]
        List of ParsedIngredient objects of structured data parsed
        from input sentences
    """
    return [parse_ingredient(sent) for sent in sentences]


@dataclass
class ParserDebugInfo:
    """Dataclass for holding intermediate objects generated during
    ingredient sentence parsing.

    Attributes
    ----------
    sentence : str
        Input ingredient sentence.
    PreProcessor : PreProcessor
        PreProcessor object created using input sentence.
    PostProcessor : PostProcessor
        PostProcessor object created using tokens, labels and scores from
        input sentence.
    Tagger : pycrfsuite.Tagger
        CRF model tagger object.
    """

    sentence: str
    PreProcessor: PreProcessor
    PostProcessor: PostProcessor
    Tagger = TAGGER


def inspect_parser(sentence: str) -> ParserDebugInfo:
    """Return object containing all intermediate objects used in the parsing of
    a sentence.

    Parameters
    ----------
    sentence : str
        Ingredient sentence to parse

    Returns
    -------
    ParserDebugInfo
        ParserDebugInfo object containing the PreProcessor object, PostProcessor
        object and Tagger.
    """
    processed_sentence = PreProcessor(sentence)
    tokens = processed_sentence.tokenized_sentence
    labels = TAGGER.tag(processed_sentence.sentence_features())
    scores = [TAGGER.marginal(label, i) for i, label in enumerate(labels)]

    # Re-plurise tokens that were singularised if the label isn't UNIT
    # For tokens with UNIT label, we'll deal with them below
    for idx in processed_sentence.singularised_indices:
        token = tokens[idx]
        label = labels[idx]
        if label != "UNIT":
            tokens[idx] = pluralise_units(token)

    postprocessed_sentence = PostProcessor(sentence, tokens, labels, scores)

    return ParserDebugInfo(
        sentence=sentence,
        PreProcessor=processed_sentence,
        PostProcessor=postprocessed_sentence,
    )



from ingredient_parser._utils import pluralise_units


class TestUtils_pluralise_units:
    def test_single(self):
        """
        Each singular unit gets pluralised
        """
        assert pluralise_units("teaspoon") == "teaspoons"
        assert pluralise_units("cup") == "cups"
        assert pluralise_units("loaf") == "loaves"
        assert pluralise_units("leaf") == "leaves"
        assert pluralise_units("chunk") == "chunks"
        assert pluralise_units("Box") == "Boxes"
        assert pluralise_units("Wedge") == "Wedges"

    def test_embedded(self):
        """
        The unit embedded in each sentence gets pluralised
        """
        assert pluralise_units("2 tablespoon olive oil") == "2 tablespoons olive oil"
        assert (
            pluralise_units("3 cup (750 milliliter) milk")
            == "3 cups (750 milliliters) milk"
        )



#!/usr/bin/env python3

import collections
import os
import platform
import re
import subprocess
from importlib.resources import as_file, files
from itertools import islice
from typing import Iterator

from ._constants import UNITS


def pluralise_units(sentence: str) -> str:
    """Pluralise units in the sentence.

    Use the same UNITS dictionary as PreProcessor to make any units in sentence plural

    Parameters
    ----------
    sentence : str
        Input sentence

    Returns
    -------
    str
        Input sentence with any words in the values of UNITS replaced with their plural
        version

    Examples
    --------
    >>> pluralise_units("2 bag")
    '2 bags'

    >>> pluralise_units("13 ounce")
    '13 ounces'

    >>> pluralise_units("1.5 loaf bread")
    '1.5 loaves bread'
    """
    for plural, singular in UNITS.items():
        sentence = re.sub(rf"\b({singular})\b", f"{plural}", sentence)

    return sentence


def consume(iterator: Iterator, n: int) -> None:
    """Advance the iterator n-steps ahead. If n is none, consume entirely.
    See consume from https://docs.python.org/3/library/itertools.html#itertools-recipes

    Parameters
    ----------
    iterator : Iterator
        Iterator to advance.
    n : int
        Number of iterations to advance by.

    Examples
    --------
    >>> it = iter(range(10))
    >>> consume(it, 3)
    >>> next(it)
    3

    >>> it = iter(range(10))
    >>> consume(it, None)
    >>> next(it)
    StopIteration
    """
    if n is None:
        # Feed the entire iterator into a zero-length deque
        collections.deque(iterator, maxlen=0)
    else:
        # Advance to the empty slice starting at position n
        next(islice(iterator, n, n), None)


def show_model_card() -> None:
    """Open model card in default application."""
    with as_file(files(__package__) / "ModelCard.md") as p:
        if platform.system() == "Darwin":  # macOS
            subprocess.call(("open", p))
        elif platform.system() == "Windows":  # Windows
            os.startfile(p)
        else:  # linux variants
            subprocess.call(("xdg-open", p))



#!/usr/bin/env python3

import re
from dataclasses import dataclass
from itertools import groupby
from operator import itemgetter
from statistics import mean
from typing import Generator, Iterator

from ._utils import consume, pluralise_units

WORD_CHAR = re.compile(r"\w")

APPROXIMATE_TOKENS = ["about", "approx.", "approximately", "nearly", "roughly"]
SINGULAR_TOKENS = ["each"]


@dataclass
class _PartialIngredientAmount:
    """Dataclass for holding the information for an ingredient amount whilst it's being
    built up.

    Attributes
    ----------
    quantity : str
        Parsed ingredient quantity
    unit : list[str] | str
        Unit or unit tokens of parsed ingredient quantity
    confidence : list[float] | float
        Average confidence of all tokens or list of confidences for each token of parsed
        ingredient amount, between 0 and 1.
    APPROXIMATE : bool, optional
        When True, indicates that the amount is approximate.
        Default is False.
    SINGULAR : bool, optional
        When True, indicates if the amount refers to a singular item of the ingredient.
        Default is False.
    """

    quantity: str
    unit: list[str] | str
    confidence: list[float] | float
    APPROXIMATE: bool = False
    SINGULAR: bool = False


@dataclass
class IngredientAmount:
    """Dataclass for holding a parsed ingredient amount, comprising the following
    attributes.

    Attributes
    ----------
    quantity : str
        Parsed ingredient quantity
    unit : str
        Unit of parsed ingredient quantity
    confidence : float
        Confidence of parsed ingredient amount, between 0 and 1.
        This is the average confidence of all tokens that contribute to this object.
    APPROXIMATE : bool, optional
        When True, indicates that the amount is approximate.
        Default is False.
    SINGULAR : bool, optional
        When True, indicates if the amount refers to a singular item of the ingredient.
        Default is False.
    """

    quantity: str
    unit: str
    confidence: float
    APPROXIMATE: bool = False
    SINGULAR: bool = False


@dataclass
class IngredientText:
    """Dataclass for holding a parsed ingredient string, comprising the following
    attributes.

    Attributes
    ----------
    text : str
        Parsed text from ingredient.
        This is comprised of all tokens with the same label.
    confidence : float
        Confidence of parsed ingredient amount, between 0 and 1.
        This is the average confidence of all tokens that contribute to this object.
    """

    text: str
    confidence: float


@dataclass
class ParsedIngredient:
    """Dataclass for holding the parsed values for an input sentence.

    Attributes
    ----------

    name : IngredientText | None
        Ingredient name parsed from input sentence.
        If no ingredient name was found, this is None.
    amount : List[IngredientAmount]
        List of IngredientAmount objects, each representing a matching quantity and
        unit pair parsed from the sentence.
    comment : IngredientText | None
        Ingredient comment parsed from input sentence.
        If no ingredient comment was found, this is None.
    other : IngredientText | None
        Any input sentence tokens that were labelled as OTHER.
        If no tokens were labelled as OTHER, this is None.
    sentence : str
        Normalised input sentence
    """

    name: IngredientText | None
    amount: list[IngredientAmount]
    comment: IngredientText | None
    other: IngredientText | None
    sentence: str


class PostProcessor:

    """Recipe ingredient sentence PostProcessor class.

    Performs the necessary postprocessing on the sentence tokens and labels and scores
    for the tokens after tagging with the CRF model in order to return a coherent
    structure of parsed information.

    Attributes
    ----------
    labels : list[str]
        List of labels for tokens.
    scores : list[float]
        Confidence associated with the label for each token.
    sentence : str
        Original ingredient sentence.
    tokens : list[str]
        List of tokens for original ingredient sentence.
    """

    def __init__(
        self, sentence: str, tokens: list[str], labels: list[str], scores: list[float]
    ):
        self.sentence = sentence
        self.tokens = tokens
        self.labels = labels
        self.scores = scores

    def __repr__(self) -> str:
        """__repr__ method

        Returns
        -------
        str
            String representation of initialised object
        """
        return f'PostProcessor("{self.sentence}")'

    def __str__(self) -> str:
        """__str__ method

        Returns
        -------
        str
            Human readable string representation of object
        """
        _str = [
            "Post-processed recipe ingredient sentence",
            f"\t{list(zip(self.tokens, self.labels))}",
        ]
        return "\n".join(_str)

    def parsed(self) -> ParsedIngredient:
        """Return parsed ingredient data

        Returns
        -------
        ParsedIngredient
            Object containing structured data from sentence.
        """
        amounts = self._postprocess_amounts()
        name = self._postprocess("NAME")
        comment = self._postprocess("COMMENT")
        other = self._postprocess("OTHER")

        return ParsedIngredient(
            sentence=self.sentence,
            amount=amounts,
            name=name,
            comment=comment,
            other=other,
        )

    def _postprocess(self, selected: str) -> IngredientText | None:
        """Process tokens, labels and scores with selected label into an
        IngredientText object.

        Parameters
        ----------
        selected : str
            Label of tokens to postprocess

        Returns
        -------
        IngredientText
            Object containing ingredient comment text and confidence
        """
        # Select indices of tokens, labels and scores for selected label
        idx = [i for i, label in enumerate(self.labels) if label == selected]

        # Join consecutive tokens together and average their score
        parts = []
        confidence_parts = []
        for group in self._group_consecutive_idx(idx):
            idx = list(group)
            joined = " ".join([self.tokens[i] for i in idx])
            confidence = mean([self.scores[i] for i in idx])

            parts.append(joined)
            confidence_parts.append(confidence)

        # Find the indices of the joined tokens list where the element
        # if a single punctuation mark or is the same as the previous element
        # in the list
        keep_idx = self._remove_isolated_punctuation_and_duplicate_indices(parts)
        parts = [parts[i] for i in keep_idx]
        confidence_parts = [confidence_parts[i] for i in keep_idx]

        # Join all the parts together into a single string and fix any
        # punctuation weirdness as a result.
        text = ", ".join(parts)
        text = self._fix_punctuation(text)

        if len(parts) == 0:
            return None

        return IngredientText(
            text=text,
            confidence=round(mean(confidence_parts), 6),
        )

    def _postprocess_amounts(self) -> list[IngredientAmount]:
        """Process tokens, labels and scores into IngredientAmount objects, by combining
        QTY labels with any following UNIT labels, up to the next QTY label.

        The confidence is the average confidence of all labels in the IngredientGroup.

        If the sequence of QTY and UNIT labels matches the "sizable unit" pattern,
        determine the amounts in a different way.

        Returns
        -------
        list[IngredientAmount]
            List of IngredientAmount objects
        """
        if match := self._sizable_unit_pattern():
            return match
        else:
            return self._fallback_pattern(self.tokens, self.labels, self.scores)

    def _fix_punctuation(self, text: str) -> str:
        """Fix some common punctuation errors that result from combining tokens of the
        same label together.

        Parameters
        ----------
        text : str
            Text resulting from combining tokens with same label

        Returns
        -------
        str
            Text, with punctuation errors fixed

        Examples
        --------
        >>> p = PostProcessor("", [], [], [])
        >>> p._fix_punctuation(", some words ( inside ),")
        "some words (inside)"

        >>> p = PostProcessor("", [], [], [])
        >>> p._fix_punctuation("(unmatched parenthesis (inside)(")
        "unmatched parenthesis (inside)"
        """
        # Remove leading comma
        if text.startswith(", "):
            text = text[2:]

        # Remove trailing comma
        if text.endswith(","):
            text = text[:-1]

        # Correct space following open parens or before close parens
        text = text.replace("( ", "(").replace(" )", ")")

        # Remove parentheses that aren't part of a matching pair
        idx_to_remove = []
        stack = []
        for i, char in enumerate(text):
            if char == "(":
                # Add index to stack when we find an opening parens
                stack.append(i)
            elif char == ")":
                if len(stack) == 0:
                    # If the stack is empty, we've found a dangling closing parens
                    idx_to_remove.append(i)
                else:
                    # Remove last added index from stack when we find a closing parens
                    stack.pop()

        # Insert anything left in stack into idx_to_remove
        idx_to_remove.extend(stack)
        text = "".join(char for i, char in enumerate(text) if i not in idx_to_remove)

        return text

    def _remove_isolated_punctuation_and_duplicate_indices(
        self, parts: list[str]
    ) -> list[int]:
        """Find elements in list that comprise a single punctuation character or are a
        duplicate of the previous element and discard their indices.

        Parameters
        ----------
        parts : list[str]
            List of tokens with single label, grouped if consecutive

        Returns
        -------
        list[int]
            Indices of elements in parts to keep

        Examples
        --------
        >>> p = PostProcessor("", [], [], [])
        >>> p._remove_isolated_punctuation_and_duplicate_indices(
            ["word", ",", "with, comma"],
        )
        [0, 2]

        >>> p = PostProcessor("", [], [], [])
        >>> p._remove_isolated_punctuation_and_duplicate_indices(
            ["word", "word", "another"],
        )
        [0, 2]
        """
        # Only keep a part if contains a word character
        idx_to_keep = []
        for i, part in enumerate(parts):
            if i == 0 and WORD_CHAR.search(part):
                idx_to_keep.append(i)
            elif WORD_CHAR.search(part) and part != parts[i - 1]:
                idx_to_keep.append(i)

        return idx_to_keep

    def _group_consecutive_idx(
        self, idx: list[int]
    ) -> Generator[Iterator[int], None, None]:
        """Yield groups of consecutive indices

        Given a list of integers, yield groups of integers where the value of each in a
        group is adjacent to the previous element's value.

        Parameters
        ----------
        idx : list[int]
            List of indices

        Yields
        ------
        list[list[int]]
            List of lists, where each sub-list contains consecutive indices

        Examples
        --------
        >>> groups = group_consecutive_idx([0, 1, 2, 4, 5, 6, 8, 9])
        >>> [list(g) for g in groups]
        [[0, 1, 2], [4, 5, 6], [8, 9]]
        """
        for k, g in groupby(enumerate(idx), key=lambda x: x[0] - x[1]):
            yield map(itemgetter(1), g)

    def _sizable_unit_pattern(self) -> list[IngredientAmount] | None:
        """Identify sentences which match the pattern where there is a
        quantity-unit pair split by one or more quantity-unit pairs e.g.

        * 1 28 ounce can
        * 2 17.3 oz (484g) package

        Return the correct sets of quantities and units, or an empty list.

        For example, for the sentence: 1 28 ounce can; the correct amounts are:
        [
            IngredientAmount(quantity="1", unit="can", score=0.x...),
            IngredientAmount(quantity="28", unit="ounce", score=0.x...),
        ]

        Returns
        -------
        list[IngredientAmount]
            List of IngredientAmount objects
        """
        # We assume that the pattern will not be longer than the first element
        # defined in patterns.
        patterns = [
            ["QTY", "QTY", "UNIT", "QTY", "UNIT", "QTY", "UNIT", "UNIT"],
            ["QTY", "QTY", "UNIT", "QTY", "UNIT", "UNIT"],
            ["QTY", "QTY", "UNIT", "UNIT"],
        ]

        # List of possible units at end of pattern that constitute a match
        end_units = [
            "bag",
            "block",
            "box",
            "can",
            "envelope",
            "jar",
            "package",
            "packet",
            "piece",
            "tin",
        ]

        amounts = []
        matching_indices = []
        for pattern in patterns:
            for match in self._match_pattern(pattern):
                # If the pattern ends with one of end_units, we have found a match for
                # this pattern!
                if self.tokens[match[-1]] in end_units:
                    # Get tokens and scores that are part of match
                    matching_tokens = [self.tokens[i] for i in match]
                    matching_scores = [self.scores[i] for i in match]

                    # Keep track of indices of matching elements so we can
                    # remove them later
                    matching_indices.extend(match)

                    # The first amount is the first and last items
                    # Note that this cannot be singular, but may be approximate
                    first = IngredientAmount(
                        quantity=matching_tokens.pop(0),
                        unit=matching_tokens.pop(-1),
                        confidence=mean(
                            [matching_scores.pop(0), matching_scores.pop(-1)]
                        ),
                        APPROXIMATE=self._is_approximate(
                            match[0], self.tokens, self.labels
                        ),
                    )
                    amounts.append(first)

                    # And create the IngredientAmount object for the pairs in between
                    for i in range(0, len(matching_tokens), 2):
                        quantity = matching_tokens[i]
                        unit = matching_tokens[i + 1]
                        confidence = mean(matching_scores[i : i + 1])

                        amount = IngredientAmount(
                            quantity=quantity,
                            unit=unit,
                            confidence=confidence,
                            SINGULAR=True,
                        )
                        amounts.append(amount)

        # If we haven't found any matches so far, return None so consumers
        # of the output of this function know there was no match.
        if len(amounts) == 0:
            return None

        # Make units plural if appropriate
        for amount in amounts:
            if amount.quantity != "1" and amount.quantity != "":
                amount.unit = pluralise_units(amount.unit)

        # Mop up any remaining amounts that didn't fit the pattern
        tokens = [tkn for i, tkn in enumerate(self.tokens) if i not in matching_indices]
        labels = [lbl for i, lbl in enumerate(self.labels) if i not in matching_indices]
        scores = [scr for i, scr in enumerate(self.scores) if i not in matching_indices]
        idx = [i for i, _ in enumerate(self.tokens) if i not in matching_indices]

        # Have a guess at where to insert them so they are in the order they appear in
        # the sentence.
        if tokens != [] and idx[0] < match[0]:
            return self._fallback_pattern(tokens, labels, scores) + amounts
        else:
            return amounts + self._fallback_pattern(tokens, labels, scores)

    def _match_pattern(self, pattern: list[str]) -> list[list[int]]:
        """Find a pattern of labels and return the indices of the labels that match the
        pattern. The pattern matching ignores labels that are not part of the pattern.

        For example, consider the sentence:
        One 15-ounce can diced tomatoes, with liquid

        It has the tokens and labels:
        ['1', '15', 'ounce', 'can', 'diced', 'tomatoes', ',', 'with', 'liquid']
        ['QTY', 'QTY', 'UNIT', 'UNIT', 'COMMENT', 'NAME', 'COMMA', 'COMMENT', 'COMMENT']

        If we search for the pattern:
        ["QTY", "QTY", "UNIT", "UNIT"]

        Then we get:
        [[0, 1, 2, 3]]

        Parameters
        ----------
        pattern : list[str]
            Pattern to match inside labels.

        Returns
        -------
        list[list[int]]
            List of label index lists that match the pattern.
        """

        if len(pattern) > len(self.labels):
            # We can never find a match.
            return []

        plen = len(pattern)
        plabels = set(pattern)

        # Select just the labels and indices of labels that are in the pattern.
        labels = [label for label in self.labels if label in plabels]
        idx = [i for i, label in enumerate(self.labels) if label in plabels]

        matches = []
        indices = iter(range(len(labels)))
        for i in indices:
            # Short circuit: If the label[i] is not equal to the first element
            # of pattern skip to next iteration
            if labels[i] == pattern[0] and labels[i : i + plen] == pattern:
                matches.append(idx[i : i + plen])
                # Advance iterator to prevent overlapping matches
                consume(indices, plen)

        return matches

    def _fallback_pattern(
        self, tokens: list[str], labels: list[str], scores: list[float]
    ) -> list[IngredientAmount]:
        """Fallback pattern for grouping quantities and units into amounts.
        This is done simply by grouping a QTY with all following UNIT until
        the next QTY.

        A special case is the for when the token "dozen" is labelled as QTY and
        it follows a QTY. In this case, the quantity of previous amount is
        modified to include "dozen".

        Parameters
        ----------
        tokens : list[str]
            Tokens for input sentence
        labels : list[str]
            Labels for input sentence tokens
        scores : list[float]
            Scores for each label

        Returns
        -------
        list[IngredientAmount]
            List of IngredientAmount objects
        """
        amounts = []
        for i, (token, label, score) in enumerate(zip(tokens, labels, scores)):
            if label == "QTY":
                # Whenever we come across a new QTY, create new IngredientAmount,
                # unless the token is "dozen" and the previous label was QTY, in which
                # case we combine modify the quantity of the previous amount.
                if token == "dozen" and labels[i - 1] == "QTY":
                    amounts[-1].quantity = amounts[-1].quantity + " dozen"
                    amounts[-1].confidence.append(score)
                else:
                    amounts.append(
                        _PartialIngredientAmount(
                            quantity=token, unit=[], confidence=[score]
                        )
                    )

            if label == "UNIT":
                if amounts == []:
                    # Not come across a QTY yet, so create IngredientAmount
                    # with no quantity
                    amounts.append(
                        _PartialIngredientAmount(
                            quantity="", unit=[], confidence=[score]
                        )
                    )

                if i > 0 and labels[i - 1] == "COMMA":
                    # If previous token was a comma, append to unit
                    # of last IngredientAmount
                    amounts[-1].unit.append(",")

                # Append token and score for unit to last IngredientAmount
                amounts[-1].unit.append(token)
                amounts[-1].confidence.append(score)

            # Check if any flags should be set
            if self._is_approximate(i, tokens, labels):
                amounts[-1].APPROXIMATE = True

            if self._is_singular(i, tokens, labels):
                amounts[-1].SINGULAR = True

            if self._is_singular_and_approximate(i, tokens, labels):
                amounts[-1].APPROXIMATE = True
                amounts[-1].SINGULAR = True

        # Loop through amounts list to fix unit and confidence
        # Unit needs converting to a string and making plural if appropriate
        # Confidence needs averaging
        # Then convert to IngredientAmount object
        processed_amounts = []
        for amount in amounts:
            combined_unit = " ".join(amount.unit)
            # Pluralise the units if appropriate
            if amount.quantity != "1" and amount.quantity != "":
                combined_unit = pluralise_units(combined_unit)

            amount.unit = combined_unit
            amount.confidence = round(mean(amount.confidence), 6)

            # Convert to an IngredientAmount object for returning
            processed_amounts.append(IngredientAmount(**amount.__dict__))

        return processed_amounts

    def _is_approximate(self, i: int, tokens: list[str], labels: list[str]) -> bool:
        """Return True is token at current index is approximate, determined
        by the token label being QTY and the previous token being in a list of
        approximate tokens.

        Parameters
        ----------
        i : int
            Index of current token
        tokens : list[str]
            List of all tokens
        labels : list[str]
            List of all token labels

        Returns
        -------
        bool
            True if current token is approximate

        Examples
        --------
        >>> p = PostProcessor("", [], [], [])
        >>> p._is_approximate(1, ["about", "3", "cups"], ["COMMENT", "QTY", "UNIT"])
        True

        >>> p = PostProcessor("", [], [], [])
        >>> p._is_approximate(1, ["approx.", "250", "g"], ["COMMENT", "QTY", "UNIT"])
        True
        """
        if i == 0:
            return False

        if labels[i] == "QTY" and tokens[i - 1].lower() in APPROXIMATE_TOKENS:
            return True

        return False

    def _is_singular(self, i: int, tokens: list[str], labels: list[str]) -> bool:
        """Return True is token at current index is singular, determined
        by the token label being UNIT and the next token being in a list of
        singular tokens.

        Parameters
        ----------
        i : int
            Index of current token
        tokens : list[str]
            List of all tokens
        labels : list[str]
            List of all token labels

        Returns
        -------
        bool
            True if current token is singular

        Examples
        --------
        >>> p = PostProcessor("", [], [], [])
        >>> p._is_approximate(1, [3", "oz", "each"], ["QTY", "UNIT", "COMMENT"])
        True
        """
        if i == len(tokens) - 1:
            return False

        if labels[i] == "UNIT" and tokens[i + 1].lower() in SINGULAR_TOKENS:
            return True

        return False

    def _is_singular_and_approximate(
        self, i: int, tokens: list[str], labels: list[str]
    ) -> bool:
        """Return True if the current token is approximate and singular, determined
        by the token label being QTY and is preceded by a token in a list of singular
        tokens, then token in a list of approximate tokens.

        e.g. each nearly 3 ...

        Parameters
        ----------
        i : int
            Index of current token
        tokens : list[str]
            List of all tokens
        labels : list[str]
            List of all token labels

        Returns
        -------
        bool
            True if current token is singular

        Examples
        --------
        >>> p = PostProcessor("", [], [], [])
        >>> p._is_approximate(
            1,
            ["nearly", "3", "oz", "each"],
            ["COMMENT", "QTY", "UNIT", "COMMENT"],
        )
        True
        """
        if i < 2:
            return False

        if (
            labels[i] == "QTY"
            and tokens[i - 1].lower() in APPROXIMATE_TOKENS
            and tokens[i - 2].lower() in SINGULAR_TOKENS
        ):
            return True

        return False



import pytest

from ingredient_parser import PostProcessor
from ingredient_parser.postprocess import (
    IngredientAmount,
    IngredientText,
    ParsedIngredient,
    consume,
)


@pytest.fixture
def p():
    """Define a PostProcessor object to use for testing the PostProcessor
    class methods.
    """
    sentence = "2 14 ounce cans coconut milk"
    tokens = ["2", "14", "ounce", "can", "coconut", "milk"]
    labels = ["QTY", "QTY", "UNIT", "UNIT", "NAME", "NAME"]
    scores = [
        0.9991370577083561,
        0.9725378063405858,
        0.9978510889596651,
        0.9922350007952175,
        0.9886087821704076,
        0.9969237827902526,
    ]

    return PostProcessor(sentence, tokens, labels, scores)


class Test_consume:
    def test_conume(self):
        """
        Test iterator advances by specified amount
        """
        it = iter(range(0, 10))
        assert next(it) == 0
        consume(it, 2)
        assert next(it) == 3

    def test_consume_all(self):
        """
        Test iterator is consumed completely
        """
        it = iter(range(0, 10))
        assert next(it) == 0
        consume(it, None)
        with pytest.raises(StopIteration):
            next(it)


class TestPostProcessor__builtins__:
    def test__str__(self, p):
        """
        Test PostProcessor __str__
        """
        truth = """Post-processed recipe ingredient sentence
\t[('2', 'QTY'), ('14', 'QTY'), ('ounce', 'UNIT'), ('can', 'UNIT'), \
('coconut', 'NAME'), ('milk', 'NAME')]"""
        assert str(p) == truth

    def test__repr__(self, p):
        """
        Test PostProessor __repr__
        """
        assert repr(p) == 'PostProcessor("2 14 ounce cans coconut milk")'


class TestPostProcessor_parsed:
    def test(self, p):
        """
        Test fixture returns expected ParsedIngredient object
        """
        expected = ParsedIngredient(
            name=IngredientText(text="coconut milk", confidence=0.992766),
            amount=[
                IngredientAmount(
                    quantity="2",
                    unit="cans",
                    confidence=0.9956860292517868,
                    APPROXIMATE=False,
                    SINGULAR=False,
                ),
                IngredientAmount(
                    quantity="14",
                    unit="ounces",
                    confidence=0.9725378063405858,
                    APPROXIMATE=False,
                    SINGULAR=True,
                ),
            ],
            comment=None,
            other=None,
            sentence="2 14 ounce cans coconut milk",
        )
        assert p.parsed() == expected


class TestPostProcessor_fix_punctuation:
    def test_leading_comma(self, p):
        """
        Test leading comma and space are removed
        """
        input_sentence = ", finely chopped"
        assert p._fix_punctuation(input_sentence) == "finely chopped"

    def test_trailing_comma(self, p):
        """
        Test trailing comma is removed
        """
        input_sentence = "finely chopped,"
        assert p._fix_punctuation(input_sentence) == "finely chopped"

    def test_space_following_open_parens(self, p):
        """
        Test space following open parenthesis is removed
        """
        input_sentence = "finely chopped ( diced)"
        assert p._fix_punctuation(input_sentence) == "finely chopped (diced)"

    def test_space_leading_close_parens(self, p):
        """
        Test space before close parenthesis is removed
        """
        input_sentence = "finely chopped (diced )"
        assert p._fix_punctuation(input_sentence) == "finely chopped (diced)"

    def test_unpaired_open_parenthesis(self, p):
        """
        Test unpaired open parenthesis is removed
        """
        input_sentence = "finely chopped diced)"
        assert p._fix_punctuation(input_sentence) == "finely chopped diced"

    def test_unpaired_close_parenthesis(self, p):
        """
        Test unpaired close parenthesis is removed
        """
        input_sentence = "finely chopped (diced"
        assert p._fix_punctuation(input_sentence) == "finely chopped diced"

    def test_multiple_unpaired_parentheses(self, p):
        """
        Test unmatched open and unmatched close parentheses are removed, but
        matched pair are kept.
        """
        input_sentence = "finely) (chopped) (diced"
        assert p._fix_punctuation(input_sentence) == "finely (chopped) diced"


class TestPostProcessor_remove_isolated_punctuation_and_duplicate_indices:
    def test_isolated_punctuation(self, p):
        """
        Test that index of "(" element is not returned
        """
        input_list = ["finely", "(", "chopped"]
        assert p._remove_isolated_punctuation_and_duplicate_indices(input_list) == [
            0,
            2,
        ]

    def test_no_isolated_punctuation(self, p):
        """
        Test all indices are returned
        """
        input_list = ["finely", "chopped", "or", "diced"]
        assert p._remove_isolated_punctuation_and_duplicate_indices(input_list) == [
            0,
            1,
            2,
            3,
        ]

    def test_adjacent_duplicate(self, p):
        """
        Test that index of second "finely" is not returned
        """
        input_list = ["finely", "finely", "chopped"]
        assert p._remove_isolated_punctuation_and_duplicate_indices(input_list) == [
            0,
            2,
        ]

    def test_non_adjacent_duplicate(self, p):
        """
        Test that index of non-adjacent duplicate is returned
        """
        input_list = ["finely", "chopped", "finely"]
        assert p._remove_isolated_punctuation_and_duplicate_indices(input_list) == [
            0,
            1,
            2,
        ]

    def test_isolated_punc_and_duplicates(self, p):
        """
        Test that index of "(" and second "finely" elements are not returned
        """
        input_list = ["finely", "finely", "(", "chopped"]
        assert p._remove_isolated_punctuation_and_duplicate_indices(input_list) == [
            0,
            3,
        ]


class TestUtils_group_consecutive_indices:
    def test_single_group(self, p):
        """
        Return single group
        """
        input_indices = [0, 1, 2, 3, 4]
        groups = p._group_consecutive_idx(input_indices)
        assert [list(g) for g in groups] == [input_indices]

    def test_multiple_groups(self, p):
        """
        Return groups of consecutive indices
        """
        input_indices = [0, 1, 2, 4, 5, 6, 8, 9]
        groups = p._group_consecutive_idx(input_indices)
        assert [list(g) for g in groups] == [[0, 1, 2], [4, 5, 6], [8, 9]]


class TestPostProcessor_sizable_unit_pattern:
    def test_long_pattern(self):
        """
        Test that 4 quantity and unit amounts are returned, with the first
        made up of the first quantity and last unit.
        """
        sentence = "1 28 ounce (400 g / 2 cups) can chickpeas"
        tokens = [
            "1",
            "28",
            "ounce",
            "(",
            "400",
            "g",
            "/",
            "2",
            "cup",
            ")",
            "can",
            "chickpeas",
        ]
        labels = [
            "QTY",
            "QTY",
            "UNIT",
            "COMMENT",
            "QTY",
            "UNIT",
            "COMMENT",
            "QTY",
            "UNIT",
            "COMMENT",
            "UNIT",
            "NAME",
        ]
        scores = [0] * len(tokens)
        p = PostProcessor(sentence, tokens, labels, scores)

        expected = [
            IngredientAmount(quantity="1", unit="can", confidence=0),
            IngredientAmount(quantity="28", unit="ounces", confidence=0, SINGULAR=True),
            IngredientAmount(quantity="400", unit="g", confidence=0, SINGULAR=True),
            IngredientAmount(quantity="2", unit="cups", confidence=0, SINGULAR=True),
        ]

        # Don't check scores
        assert len(p._sizable_unit_pattern()) == len(expected)
        for output, expected in zip(p._sizable_unit_pattern(), expected):
            assert output.quantity == expected.quantity
            assert output.unit == expected.unit
            assert output.SINGULAR == expected.SINGULAR
            assert output.APPROXIMATE == expected.APPROXIMATE

    def test_medium_pattern(self):
        """
        Test that 3 quantity and unit amounts are returned, with the first
        made up of the first quantity and last unit.
        """
        sentence = "1 28 ounce (400 g) can chickpeas"
        tokens = [
            "1",
            "28",
            "ounce",
            "(",
            "400",
            "g",
            ")",
            "can",
            "chickpeas",
        ]
        labels = [
            "QTY",
            "QTY",
            "UNIT",
            "COMMENT",
            "QTY",
            "UNIT",
            "COMMENT",
            "UNIT",
            "NAME",
        ]
        scores = [0] * len(tokens)
        p = PostProcessor(sentence, tokens, labels, scores)

        expected = [
            IngredientAmount(quantity="1", unit="can", confidence=0),
            IngredientAmount(quantity="28", unit="ounces", confidence=0, SINGULAR=True),
            IngredientAmount(quantity="400", unit="g", confidence=0, SINGULAR=True),
        ]

        # Don't check scores
        assert len(p._sizable_unit_pattern()) == len(expected)
        for output, expected in zip(p._sizable_unit_pattern(), expected):
            assert output.quantity == expected.quantity
            assert output.unit == expected.unit
            assert output.SINGULAR == expected.SINGULAR
            assert output.APPROXIMATE == expected.APPROXIMATE

    def test_short_pattern(self):
        """
        Test that 4 quantity and unit amounts are returned, with the first
        made up of the first quantity and last unit.
        """
        sentence = "1 28 ounce can chickpeas"
        tokens = [
            "1",
            "28",
            "ounce",
            "can",
            "chickpeas",
        ]
        labels = [
            "QTY",
            "QTY",
            "UNIT",
            "UNIT",
            "NAME",
        ]
        scores = [0] * len(tokens)
        p = PostProcessor(sentence, tokens, labels, scores)

        expected = [
            IngredientAmount(quantity="1", unit="can", confidence=0),
            IngredientAmount(quantity="28", unit="ounces", confidence=0, SINGULAR=True),
        ]

        # Don't check scores
        assert len(p._sizable_unit_pattern()) == len(expected)
        for output, expected in zip(p._sizable_unit_pattern(), expected):
            assert output.quantity == expected.quantity
            assert output.unit == expected.unit
            assert output.SINGULAR == expected.SINGULAR
            assert output.APPROXIMATE == expected.APPROXIMATE

    def test_no_pattern(self):
        """
        Test that None is return if pattern is not matched
        """
        sentence = "400 g chickpeas or black beans"
        tokens = ["400", "g", "chickpeas", "or", "black", "beans"]
        labels = ["QTY", "UNIT", "NAME", "NAME", "NAME", "NAME"]
        scores = [0] * len(tokens)
        p = PostProcessor(sentence, tokens, labels, scores)

        # Don't check scores
        assert p._sizable_unit_pattern() is None

    def test_mixed_pattern(self):
        """
        Test that 3 quantity and unit amounts are returned, with the first
        made up of the first quantity and last unit.
        """
        sentence = "2 cups or 1 28 ounce can chickpeas"
        tokens = ["2", "cup", "or", "1", "28", "ounce", "can", "chickpeas"]
        labels = ["QTY", "UNIT", "COMMENT", "QTY", "QTY", "UNIT", "UNIT", "NAME"]
        scores = [0] * len(tokens)
        p = PostProcessor(sentence, tokens, labels, scores)

        expected = [
            IngredientAmount(quantity="2", unit="cups", confidence=0),
            IngredientAmount(quantity="1", unit="can", confidence=0),
            IngredientAmount(quantity="28", unit="ounces", confidence=0, SINGULAR=True),
        ]

        # Don't check scores
        assert len(p._sizable_unit_pattern()) == len(expected)
        for output, expected in zip(p._sizable_unit_pattern(), expected):
            assert output.quantity == expected.quantity
            assert output.unit == expected.unit
            assert output.SINGULAR == expected.SINGULAR
            assert output.APPROXIMATE == expected.APPROXIMATE


class TestPostProcessor_match_pattern:
    def test_long_pattern_match(self):
        """
        Test that correct start and stop indices are returned for long pattern
        """
        pattern = ["QTY", "QTY", "UNIT", "QTY", "UNIT", "QTY", "UNIT", "UNIT"]

        labels = [
            "QTY",
            "UNIT",
            "QTY",
            "QTY",
            "UNIT",
            "QTY",
            "UNIT",
            "QTY",
            "UNIT",
            "UNIT",
        ]
        p = PostProcessor("", [], labels, [])

        assert p._match_pattern(pattern) == [[2, 3, 4, 5, 6, 7, 8, 9]]

    def test_medium_pattern_match(self):
        """
        Test that correct start and stop indices are returned for medium pattern
        """
        pattern = ["QTY", "QTY", "UNIT", "QTY", "UNIT", "UNIT"]

        labels = [
            "QTY",
            "QTY",
            "UNIT",
            "QTY",
            "UNIT",
            "UNIT",
            "UNIT",
        ]
        p = PostProcessor("", [], labels, [])

        assert p._match_pattern(pattern) == [[0, 1, 2, 3, 4, 5]]

    def test_short_pattern_match(self):
        """
        Test that correct start and stop indices are returned for long pattern
        """
        pattern = ["QTY", "QTY", "UNIT", "UNIT"]

        labels = [
            "QTY",
            "UNIT",
            "QTY",
            "QTY",
            "UNIT",
            "UNIT",
            "QTY",
            "UNIT",
            "UNIT",
        ]
        p = PostProcessor("", [], labels, [])

        assert p._match_pattern(pattern) == [[2, 3, 4, 5]]

    def test_impossible_match(self):
        """
        Test that empty list is returned when match is impossible beacause pattern
        is longer than list of labels
        """
        pattern = ["QTY", "QTY", "UNIT", "QTY", "UNIT", "UNIT"]

        labels = [
            "QTY",
            "QTY",
            "UNIT",
            "UNIT",
        ]
        p = PostProcessor("", [], labels, [])

        assert p._match_pattern(pattern) == []

    def test_multiple_matches(self):
        """
        Test that multiple non-overlapping matches are returned
        """
        pattern = ["QTY", "QTY", "UNIT", "UNIT"]

        labels = [
            "QTY",
            "QTY",
            "UNIT",
            "UNIT",
            "QTY",
            "QTY",
            "QTY",
            "UNIT",
            "UNIT",
        ]
        p = PostProcessor("", [], labels, [])

        assert p._match_pattern(pattern) == [[0, 1, 2, 3], [5, 6, 7, 8]]


class TestPostProcessor_fallback_pattern:
    def test_basic(self, p):
        """
        Test that a single IngredientAmount object with quantity "3" and
        unit "large handfuls" is returned.
        """

        tokens = ["3", "large", "handful", "cherry", "tomatoes"]
        labels = ["QTY", "UNIT", "UNIT", "NAME", "NAME"]
        scores = [0] * len(tokens)

        expected = [IngredientAmount(quantity="3", unit="large handfuls", confidence=0)]

        assert p._fallback_pattern(tokens, labels, scores) == expected

    def test_no_quantity(self, p):
        """
        Test that a single IngredientAmount object with no quantity and
        unit "large handfuls" is returned.
        """

        tokens = ["1", "green", ",", "large", "pepper"]
        labels = ["QTY", "NAME", "COMMA", "UNIT", "NAME"]
        scores = [0] * len(tokens)

        expected = [IngredientAmount(quantity="1", unit=", large", confidence=0)]

        assert p._fallback_pattern(tokens, labels, scores) == expected

    def test_comma_before_unit(self, p):
        """
        Test that a single IngredientAmount object with no quantity and
        unit "large handfuls" is returned.
        """

        tokens = ["bunch", "of", "basil", "leaves"]
        labels = ["UNIT", "COMMENT", "NAME", "NAME"]
        scores = [0] * len(tokens)

        expected = [IngredientAmount(quantity="", unit="bunch", confidence=0)]

        assert p._fallback_pattern(tokens, labels, scores) == expected

    def test_approximate(self, p):
        """
        Test that a single IngredientAmount object with the APPROXIMATE flag set
        is returned
        """
        tokens = ["About", "2", "cup", "flour"]
        labels = ["COMMENT", "QTY", "UNIT", "NAME"]
        scores = [0] * len(tokens)

        expected = [
            IngredientAmount(
                quantity="2",
                unit="cups",
                confidence=0,
                APPROXIMATE=True,
            )
        ]

        assert p._fallback_pattern(tokens, labels, scores) == expected

    def test_singular(self, p):
        """
        Test that a single IngredientAmount object with the SINGULAR flag set
        is returned
        """
        tokens = ["2", "bananas", ",", "4", "ounce", "each"]
        labels = ["QTY", "NAME", "COMMA", "QTY", "UNIT", "COMMENT"]
        scores = [0] * len(tokens)

        expected = [
            IngredientAmount(
                quantity="2",
                unit="",
                confidence=0,
            ),
            IngredientAmount(
                quantity="4",
                unit="ounces",
                confidence=0,
                SINGULAR=True,
            ),
        ]

        assert p._fallback_pattern(tokens, labels, scores) == expected

    def test_singular_and_approximate(self, p):
        """
        Test that a single IngredientAmount object with the APPROXIMATE and
        SINGULAR flags set is returned
        """
        tokens = ["2", "bananas", ",", "each", "about", "4", "ounce"]
        labels = ["QTY", "NAME", "COMMA", "COMMENT", "COMMENT", "QTY", "UNIT"]
        scores = [0] * len(tokens)

        expected = [
            IngredientAmount(
                quantity="2",
                unit="",
                confidence=0,
            ),
            IngredientAmount(
                quantity="4",
                unit="ounces",
                confidence=0,
                SINGULAR=True,
                APPROXIMATE=True,
            ),
        ]

        assert p._fallback_pattern(tokens, labels, scores) == expected

    def test_dozen(self, p):
        """
        Test that the token "dozen" is combined with the preceding QTY token in a
        single IngredientAmount object.
        """
        tokens = ["2", "dozen", "bananas", ",", "each", "about", "4", "ounce"]
        labels = ["QTY", "QTY", "NAME", "COMMA", "COMMENT", "COMMENT", "QTY", "UNIT"]
        scores = [0] * len(tokens)

        expected = [
            IngredientAmount(
                quantity="2 dozen",
                unit="",
                confidence=0,
            ),
            IngredientAmount(
                quantity="4",
                unit="ounces",
                confidence=0,
                SINGULAR=True,
                APPROXIMATE=True,
            ),
        ]

        assert p._fallback_pattern(tokens, labels, scores) == expected


class TestPostProcessor_is_approximate:
    def test_is_approximate_about(self, p):
        """
        Test that QTY at index is indicated as approximate
        """
        tokens = ["about", "5", "cups", "orange", "juice"]
        labels = ["COMMENT", "QTY", "UNIT", "NAME", "NAME"]
        assert p._is_approximate(1, tokens, labels)

    def test_is_approximate_approx(self, p):
        """
        Test that QTY at index is indicated as approximate
        """
        tokens = ["approx.", "5", "cups", "orange", "juice"]
        labels = ["COMMENT", "QTY", "UNIT", "NAME", "NAME"]
        assert p._is_approximate(1, tokens, labels)

    def test_is_approximate_approximately(self, p):
        """
        Test that QTY at index is indicated as approximate
        """
        tokens = ["approximately", "5", "cups", "orange", "juice"]
        labels = ["COMMENT", "QTY", "UNIT", "NAME", "NAME"]
        assert p._is_approximate(1, tokens, labels)

    def test_is_approximate_nearly(self, p):
        """
        Test that QTY at index is indicated as approximate
        """
        tokens = ["nearly", "5", "cups", "orange", "juice"]
        labels = ["COMMENT", "QTY", "UNIT", "NAME", "NAME"]
        assert p._is_approximate(1, tokens, labels)

    def test_not_approximate(self, p):
        """
        Test that QTY at index is not indicated as approximate
        """
        tokens = ["maximum", "5", "cups", "orange", "juice"]
        labels = ["COMMENT", "QTY", "UNIT", "NAME", "NAME"]
        assert not p._is_approximate(1, tokens, labels)


class TestPostProcessor_is_singular:
    def test_is_singular(self, p):
        """
        Test that UNIT at index is indicated as singular
        """
        tokens = ["4", "salmon", "fillets", "2", "pounds", "each"]
        labels = ["QTY", "NAME", "NAME", "QTY", "UNIT", "COMMENT"]
        assert p._is_singular(4, tokens, labels)

    def test_not_singular(self, p):
        """
        Test that UNIT at index is not indicated as singular
        """
        tokens = ["4", "salmon", "fillets", "2", "pounds", "minimum"]
        labels = ["QTY", "NAME", "NAME", "QTY", "UNIT", "COMMENT"]
        assert not p._is_singular(4, tokens, labels)


class TestPostProcessor_is_singular_and_approximate:
    def test_is_singular_and_approximate(self, p):
        """
        Test that QTY at index is indicated as approximate and singular
        """
        tokens = ["each", "nearly", "2", "kg"]
        labels = ["COMMENT", "COMMENT", "QTY", "UNIT"]
        assert p._is_singular_and_approximate(2, tokens, labels)

    def test_not_singular_and_approximate(self, p):
        """
        Test that QTY at index is not indicated as approximate and singular
        """
        tokens = ["both", "about", "2", "kg"]
        labels = ["COMMENT", "COMMENT", "QTY", "UNIT"]
        assert not p._is_singular_and_approximate(2, tokens, labels)



import json
from itertools import zip_longest

import pytest

from ingredient_parser import parse_ingredient
from ingredient_parser.postprocess import (
    IngredientAmount,
    IngredientText,
    ParsedIngredient,
)


def dict_to_ParsedIngredient(data: dict[str, str]) -> ParsedIngredient:
    """Convert test data dictionary to ParsedIngredient object

    Parameters
    ----------
    data : dict[str, str]
        Dictionary of test data containing input sentence and true labels.

    Returns
    -------
    ParsedIngredient
        Object of sentence and true labels in same format as output from
        parse_ingredient function.
    """
    sentence = data["sentence"]
    labels = data["labels"]

    if true_name := labels.get("name", None):
        name = IngredientText(text=true_name, confidence=0)
    else:
        name = None

    if true_comment := labels.get("comment", None):
        comment = IngredientText(text=true_comment, confidence=0)
    else:
        comment = None

    amount = IngredientAmount(
        quantity=labels.get("quantity", ""),
        unit=labels.get("unit", ""),
        confidence=0,
    )

    return ParsedIngredient(
        name=name, amount=[amount], comment=comment, other=None, sentence=sentence
    )


def load_test_cases(file: str):
    """Load newline delimited json file of test cases

    Parameters
    ----------
    file : str
        Filename

    Returns
    -------
    List[Dict[str,str]]
        List of dicts of test cases.
        Each dict contains:
            sentence: the input sentence to parse
            labels: the correct labels
    """
    sample_data = []
    with open(file, "r") as f:
        test_cases = f.read().splitlines()
        for case in test_cases:
            if case.startswith("#"):
                continue
            data = json.loads(case)
            obj = dict_to_ParsedIngredient(data)
            sample_data.append(obj)

    return sample_data


class TestParser:
    @pytest.mark.parametrize("truth", load_test_cases("tests/test_parser_simple.json"))
    def test_parser_simple(self, truth):
        """Test simple cases of ingredient sentences that the parser should have no
        issues getting correct

        Parameters
        ----------
        truth : ParsedIngredient
            ParsedIngredient object containing input sentence and true parsed data
        """
        parsed = parse_ingredient(truth.sentence)

        if truth.name is not None:
            assert parsed.name.text == truth.name.text
        else:
            assert parsed.name is None

        if truth.comment is not None:
            assert parsed.comment.text == truth.comment.text
        else:
            assert parsed.comment is None

        if truth.other is not None:
            assert parsed.other.text == truth.other.text
        else:
            assert parsed.other is None

        for output, expected in zip_longest(truth.amount, parsed.amount):
            assert output.quantity == expected.quantity
            assert output.unit == expected.unit

    @pytest.mark.parametrize(
        "truth", load_test_cases("tests/test_parser_alternate_unit_position.json")
    )
    def test_parser_alternate_unit_position(self, truth):
        """Test cases where the unit appears after the ingredient name

        Parameters
        ----------
        truth : ParsedIngredient
            ParsedIngredient object containing input sentence and true parsed data
        """
        parsed = parse_ingredient(truth.sentence)

        if truth.name is not None:
            assert parsed.name.text == truth.name.text
        else:
            assert parsed.name is None

        if truth.comment is not None:
            assert parsed.comment.text == truth.comment.text
        else:
            assert parsed.comment is None

        if truth.other is not None:
            assert parsed.other.text == truth.other.text
        else:
            assert parsed.other is None

        for output, expected in zip_longest(truth.amount, parsed.amount):
            assert output.quantity == expected.quantity
            assert output.unit == expected.unit

    @pytest.mark.parametrize(
        "truth", load_test_cases("tests/test_parser_ambiguous_unit.json")
    )
    def test_parser_ambiguous_unit(self, truth):
        """Test cases where the same word could be a unit or an ingredient,
        depending on context e.g. clove

        Parameters
        ----------
        truth : ParsedIngredient
            ParsedIngredient object containing input sentence and true parsed data
        """
        parsed = parse_ingredient(truth.sentence)

        if truth.name is not None:
            assert parsed.name.text == truth.name.text
        else:
            assert parsed.name is None

        if truth.comment is not None:
            assert parsed.comment.text == truth.comment.text
        else:
            assert parsed.comment is None

        if truth.other is not None:
            assert parsed.other.text == truth.other.text
        else:
            assert parsed.other is None

        for output, expected in zip_longest(truth.amount, parsed.amount):
            assert output.quantity == expected.quantity
            assert output.unit == expected.unit


#    @pytest.mark.parametrize(
#        "data", load_test_cases("tests/test_parser_comment_before_name.json")
#    )
#    def test_parser_comment_before_name(self, data):
#        """Test cases where the comment appears before the ingredient name,
#        e.g. 1 chopped carrot
#
#        Parameters
#        ----------
#        data : Dict[str, Dict[str, str]]
#            Dictionary of sample data with the following keys
#                sentence - the input sentence
#                labels - the correct labels the parser should return
#        """
#        sentence = data["sentence"]
#        true_labels = data["labels"]
#        parsed = parse_ingredient(sentence)
#
#        # parse_ingredient returns a dict with more keys than just the labels
#        # so iterate through each label in true_labels and only check them.
#        for key in true_labels:
#            assert parsed.__dict__[key] == true_labels[key]
#
#    @pytest.mark.parametrize(
#        "data", load_test_cases("tests/test_parser_multiple_units.json")
#    )
#    def test_parser_multiple_units(self, data):
#        """Test cases where there are multiple units in the sentence
#
#        Parameters
#        ----------
#        data : Dict[str, Dict[str, str]]
#            Dictionary of sample data with the following keys
#                sentence - the input sentence
#                labels - the correct labels the parser should return
#        """
#        sentence = data["sentence"]
#        true_labels = data["labels"]
#        parsed = parse_ingredient(sentence)
#
#        # parse_ingredient returns a dict with more keys than just the labels
#        # so iterate through each label in true_labels and only check them.
#        for key in true_labels:
#            assert parsed.__dict__[key] == true_labels[key]



from ._utils import show_model_card
from .parsers import inspect_parser, parse_ingredient, parse_multiple_ingredients
from .postprocess import PostProcessor
from .preprocess import PreProcessor

__all__ = [
    "inspect_parser",
    "parse_ingredient",
    "parse_multiple_ingredients",
    "PreProcessor",
    "PostProcessor",
    "show_model_card",
]

__version__ = "0.1.0-beta5"
